#!/usr/bin/env python3


import random
import os.path
import shutil
import argparse
import subprocess
from pprint import pprint
import time
from datetime import date

# 23-02 these are all weighted MaxSAT evaluation 22 solver
# see: https://maxsat-evaluations.github.io/2022/results/complete/weighted/summary.html
# PLEASE ADD OWN SOLVERS TO THE DICTIONARY IN THE FOLLOWING WAY
# wcnf file is added automatically by the script at the end of each solver_call
# directory (.) is in relation to THIS script
# solver["YOURSOLVERNAME"] = {
#     "solver_call" : "SOLVERCALL . is the folder of the current script"}
solvers = {}
solvers["Cashwmaxsat-CorePlus"] = {
    "solver_call": "./Cashwmaxsat-CorePlus/bin/cashwmaxsatcoreplus -v0 -no-bin -no-sat -m -no-par -bm"}
solvers["Cashwmaxsat-Plus"] = {
    "solver_call": "./Cashwmaxsat-Plus/bin/cashwmaxsatplus -v0 -no-bin -no-sat -no-par -bm -m"}
solvers["UWrMaxSat-SCIP"] = {
    "solver_call": "./UWrMaxSat-SCIP/bin/uwrmaxsat -v1 -no-bin -no-sat -no-par -scip-cpu=500 -m -bm"}
solvers["MaxHS"] = {
    "solver_call": "./MaxHS/bin/maxhs -no-printOptions -printSoln -verb=0"}
solvers["WMaxCDCL"] = {
    "solver_call": "./WMaxCDCL/bin/wmaxcdcl_static"}
solvers["WMaxCDCL-BandAll"] = {
    "solver_call": "./WMaxCDCL-BandAll/bin/wmaxcdcl_static"}
solvers["UWrMaxSat"] = {
    "solver_call": "./UWrMaxSat/bin/uwrmaxsat -v1 -no-sat -no-bin -m -bm"}
solvers["EvalMaxSAT"] = {
    "solver_call": "./EvalMaxSAT/bin/EvalMaxSAT_bin"}
# ATTENTION: for CGSS python2 and the package python-six is needed
solvers["CGSS"] = {
    "solver_call": "python2 ./CGSS/bin/rc2.py --comp=a"}
solvers["Exact"] = {
    "solver_call": "./Exact/bin/Exact_e17dbb7 --verbosity=0 --format=wcnf --output=maxsatnew --bits-overflow=30 --bits-reduced=25"}
# ATTENTION: Z3 Solver calls are  more costly as the wcnf has to be transformed into the old format. This is done in the bash script.
solvers["z3maxres"] = {
    "solver_call": "./z3maxres.sh"}
solvers["z3wmax"] = {
    "solver_call": "./z3wmax.sh"}
solvers["z3rc2"] = {
    "solver_call": "./z3rc2.sh"}
solvers["z3pd-maxres"] = {
    "solver_call": "./z3pd-maxres.sh"}
solvers["z3maxres-bin"] = {
    "solver_call": "./z3maxres-bin.sh"}
# MSE21 Version
solvers["Pacose"] = {
    "solver_call": "./Pacose.sh"}
# there is a problem with the model
solvers["maxpre2EvalMaxSAT"] = {
    "solver_call": "./maxpre2.sh"}



def is_valid_file(arg):
    if not os.path.exists(arg):
        print("The file %s does not exist!" % arg)
        exit(1)


parser = argparse.ArgumentParser(
    description='This is a awesome MaxSAT Solver comparison script!!')
parser.add_argument("wcnfFile",
                    help="WCNF file which should be checked (in relation to current path).")
parser.add_argument("--logging", action="store_true",
                    help="Activates logging. A logfile for ALL SOLVER FROM SOLVERS LIST is created. (if --solver is not used this equals all solvers)")
parser.add_argument("--logAll", action="store_true",
                    help="Activates logging. A logfile for EVERY erroneous solver is created.")
parser.add_argument("--saveWCNF", action="store_true",
                    help=" Copies the WCNF to folder: (default: <date>_WCNFs/).")
parser.add_argument("--saveWCNFFolder", default=str(date.today()) + "_WCNFs",
                    help="Add error WCNFs to folder (default: <date>_WCNFs/).")
parser.add_argument("--logPath", default=str(date.today()) + "_FaultLogs",
                    help="Add logfiles for each erroneous solver in folder (default ./ErrorLogs).")
parser.add_argument("--timeout", default=20, type=int,
                    help="Timeout for each MaxSAT Solver.")
parser.add_argument('-l', '--list', help='delimited list input',
                    type=lambda s: [int(item) for item in s.split(',')])
parser.add_argument("--solvers", default=solvers.keys(), type=lambda s: [item for item in s.split(',')],
                    help='Only return an invalid return code if a solver from this list is faulty, but all solvers run to have a better mayority voting. Write them without empty space!!! --solvers a,b,c --> a,b,c in ' + str(solvers.keys()), required=False)
# wcnverifier
# rv == 1: a hard clause is not satifiable with given v-line
# rv == 2: o value is smaller than the calculated value from verifier
# rv == 3: o value is bigger than the calculated value from verifier
# rv == 11: hard clauses are satisfiable
# rv == 12: sat solver returned unknown
# rv == 42: should never happen
parser.add_argument("--verifier",  nargs='?', default="./wcnfverifier/wcnfverifier",
                    help="How to call the verifier to check models and UNSAT results (total path or in relation to this script directory).")
parser.add_argument("--satSolver", nargs='?', default="./wcnfverifier/kissat/build/kissat",
                    help="How to call the SAT solver to verify the UNSAT results -- used by verifier (total path or in relation to this script directory).")


args = parser.parse_args()

is_valid_file(str(args.wcnfFile))

for solarg in args.solvers:
    if solarg not in solvers.keys():
        print(solarg + " is not a recognized MaxSAT solver from the list:")
        print(solvers.keys())
        exit(1)

# seed to add to each file created in /tmp/, try to add seed of current filename:
seed = ""
with open(args.wcnfFile, 'r') as f:
    first_line = f.readline()
    if "c seed" in first_line:
        seed = str(first_line[7:-1])
if seed == "" and any(char.isdigit() for char in os.path.splitext(os.path.basename(args.wcnfFile))[0]):
    seed = str(int(''.join(
        filter(str.isdigit, os.path.splitext(os.path.basename(args.wcnfFile))[0]))))
if seed == "":
    seed = str(random.randrange(1, 2**32-1))
long_seed = seed + "-" + str(random.randrange(1, 2**32-1))

current_directory = os.getcwd()
current_script_directory = os.path.dirname(os.path.abspath(__file__))

if (args.verifier[0] == "."):
    args.verifier = current_script_directory + args.verifier[1:]
if (args.satSolver[0] == "."):
    args.satSolver = current_script_directory + args.satSolver[1:]
if args.logging or args.logAll:
    if not os.path.exists(args.logPath):
        os.makedirs(args.logPath)

is_valid_file(str(args.satSolver))
is_valid_file(str(args.verifier))


def SolverCall(solver):

    # if (solver["solver_call"][0] == "."):
    #     solver["solver_call"] = current_script_directory + \
    #         solver["solver_call"][1:]
    solver["solver_call"] = solver["solver_call"].replace(
        "./", current_script_directory + "/")
    # print(solver["solver_call"])
    # print(current_script_directory)

    start_time = time.time()

    solverOut = subprocess.run(
        "timeout --verbose --kill-after=10 " + str(args.timeout) + " " + solver["solver_call"] + " " + args.wcnfFile, shell=True, capture_output=True)

    if solverOut.returncode == 124:
        # print("TIMEOUT")
        solver["s-status"] = "TIMEOUT"
        solver["ver_result"] = "TIMEOUT"
        solver["time"] = time.time()-start_time
        return

    solver["time"] = time.time()-start_time
    solver["return_code"] = solverOut.returncode
    solver["stdout"] = solverOut.stdout.decode('utf-8')
    solver["stderr"] = solverOut.stderr.decode('utf-8')

    for line in solver["stdout"].split("\n"):
        # print(line)
        if line.startswith('s '):
            solver["s-status"] = line[2:]
            if solver["s-status"] == "SATISFIABLE":
                solver["s-status"] = "OPTIMUM FOUND"
        if "c SCIP result: UNSATISFIABLE" == line:
            solver["s-status"] = "UNSATISFIABLE"
        if "UNSATISFIABLE" == line:
            solver["s-status"] = "UNSATISFIABLE"
        if line.startswith('o '):
            if line[2:] != "-oo":
                solver["o-value"] = int(line[2:])
            else:
                # only hard clauses -- but they are satisfiable
                solver["o-value"] = 0
        if line.startswith('v '):
            solver["model"] = line[2:]
        if "Error" in line or "ERROR" in line or "error" in line:
            solver["stdout_error"] = line

    if "s-status" not in solver:
        solver["s-status"] = "STRANGE"

    if solver["s-status"] == "OPTIMUM FOUND":
        if "model" not in solver:
            solver["ERROR_solver"] = "opt_but_no_model"
        elif "o-value" not in solver:
            solver["ERROR_solver"] = "opt_but_no_o-value"
    else:
        if "model" in solver and "o-value" in solver:
            solver["ERROR_solver"] = "model_and_o-value_but_no_opt"

    if solver["s-status"] == "OPTIMUM FOUND" and "ERROR_solver" not in solver:
        model_file = "/tmp/stdout-" + long_seed + ".model"
        with open(model_file, "w") as text_file:
            text_file.write("v " + solver["model"] + "\n")

        verifyCall = args.verifier + " -o " + \
            str(solver["o-value"]) + " -v " + model_file + " " + args.wcnfFile
        verifierData = subprocess.run(
            verifyCall, shell=True, capture_output=True)

        solver["ver_return_code"] = verifierData.returncode
        os.remove(model_file)

        verifierDataCall = verifierData.stdout.decode('utf-8')
        for line in verifierDataCall.split("\n"):
            if line.startswith('o '):
                solver["ver_o-value"] = int(line[2:])

    if verifiedUNSAT == "UNSATISFIABLE" and solver["s-status"] == "UNSATISFIABLE":
        solver["ver_result"] = "UNSATISFIABLE"
    elif verifiedUNSAT == "SATISFIABLE" and solver["s-status"] == "OPTIMUM FOUND":
        if ("ver_o-value" in solver and "o-value" in solver and solver["ver_o-value"] == solver["o-value"]):
            solver["ver_result"] = "OPTIMUM FOUND"
        else:
            solver["ver_result"] = "FALSE VALUE"
    elif verifiedUNSAT == "UNSATISFIABLE" and solver["s-status"] == "OPTIMUM FOUND":
        solver["ver_result"] = "UNSAT BUT VALUE"
    elif verifiedUNSAT == "SATISFIABLE" and solver["s-status"] == "UNSATISFIABLE":
        solver["ver_result"] = "FALSE UNSAT"
    else:
        solver["ver_result"] = "UNKNOWN"


verifierUNSATData = subprocess.run(
    args.verifier + " -s " + args.satSolver + " " + args.wcnfFile, shell=True, capture_output=True)
if verifierUNSATData.returncode == 11:
    verifiedUNSAT = "SATISFIABLE"
elif verifierUNSATData.returncode == 0:
    verifiedUNSAT = "UNSATISFIABLE"
else:
    verifiedUNSAT = str(verifierUNSATData.returncode)

verifierUNSATCall = verifierUNSATData.stdout.decode('utf-8')
# print(verifierUNSATCall)
# import sys
# if verifiedUNSAT != "SATISFIABLE" and verifiedUNSAT != "UNSATISFIABLE":
#     print("Error message: " + verifiedUNSAT, file=sys.stderr)
#     print(verifierUNSATCall, file=sys.stderr)
#     print(verifierUNSATData.stderr.decode('utf-8'), file=sys.stderr)
instance_stats = ""
for line in verifierUNSATCall.split("\n"):
    if line.startswith('c Clauses......: '):
        instance_stats += line + "\n"
    if line.startswith('c SoftClauses..: '):
        instance_stats += line + "\n"
    if line.startswith('c Variables....: '):
        instance_stats += line + "\n"
    if line.startswith('c MaxWeight....: '):
        MaxWeight = int(line[17:])
        instance_stats += line + "\n"
    if line.startswith('c SumOfWeights.: '):
        SumOfWeights = int(line[17:])
        instance_stats += line


assert (MaxWeight < 2**63)
assert (SumOfWeights < 2**64)

minVerifiedOValue = 2**64
minValueHitBySolver = []
verifiedUNSATHitBySolver = []
solverid = 0
totalTime = 0
numberTimeouts = 0
for solver in solvers:
    solverid += 1
    # print("\n\n"+solver)
    SolverCall(solvers[solver])
    solvers[solver]["id"] = solverid
    if solvers[solver]["ver_result"] != "TIMEOUT":
        totalTime += solvers[solver]["time"]
    else:
        numberTimeouts += 1
    solvers[solver]["name"] = solver
    # print(solvers[solver])

    if solvers[solver]["ver_result"] == "OPTIMUM FOUND" and minVerifiedOValue >= solvers[solver]["o-value"] and solvers[solver]["o-value"] >= 0:
        if minVerifiedOValue > solvers[solver]["o-value"]:
            minVerifiedOValue = solvers[solver]["o-value"]
            minValueHitBySolver = []
            minValueHitBySolver.append(solver)
        else:
            minValueHitBySolver.append(solver)
    elif solvers[solver]["ver_result"] == "UNSATISFIABLE":
        verifiedUNSATHitBySolver.append(solver)
averageTime=totalTime/(solverid-numberTimeouts)

BiggerUINT32 = 0
if int(SumOfWeights) >= 2**32:
    BiggerUINT32 = 30

assert (int(SumOfWeights) < 2**64)

timeoutFactor = 50
for solver in solvers:
    if solvers[solver]["ver_result"] == "TIMEOUT":
        if numberTimeouts < len(solvers)/2 and averageTime*timeoutFactor < args.timeout:
            solvers[solver]["ver_result"] = "TIMEOUT"
        else:
            solvers[solver]["ver_result"] = "TIMEOUT -- ignored"


overall_return_code = 0
return_code = 0

faulty_solvers = []
logging_files = []
for solver in solvers:
    if solvers[solver]["ver_result"] == "TIMEOUT -- ignored":
        solvers[solver]["error_code"] = 0
    elif solvers[solver]["ver_result"] == "TIMEOUT":
        solvers[solver]["error_code"] = 1 + BiggerUINT32
        solvers[solver]["fault_description"] = "POTENTIAL ERROR: TIMEOUT is " + \
            str(timeoutFactor) + \
            " times bigger than accumulated time of all other solvers."
    elif solvers[solver].get('return_code', 0) == 1:
        solvers[solver]["error_code"] = 2 + BiggerUINT32
        solvers[solver]["fault_description"] = "Return Code of MaxSAT solver == 1"
    elif solvers[solver].get('return_code', 0) == 134:
        solvers[solver]["error_code"] = 3 + BiggerUINT32
        solvers[solver][
            "fault_description"] = "Return Code of MaxSAT solver == 134 (SIGABRT)"
    elif solvers[solver].get('return_code', 0) == 136:
        solvers[solver]["error_code"] = 4 + BiggerUINT32
        solvers[solver][
            "fault_description"] = "Return Code of MaxSAT solver == 136 (SIGTERM / SIGABORT)"
    elif solvers[solver].get('return_code', 0) == 137:
        solvers[solver]["error_code"] = 25 + BiggerUINT32
        solvers[solver][
            "fault_description"] = "Return Code of MaxSAT solver == 137 (SIGTERM / SIGABORT)"
    elif solvers[solver].get('return_code', 0) == 139:
        solvers[solver]["error_code"] = 26 + BiggerUINT32
        solvers[solver][
            "fault_description"] = "Return Code of MaxSAT solver == 139 (SIGTERM / SIGABORT)"
    elif solvers[solver].get('return_code', 0) == 255:
        solvers[solver]["error_code"] = 27 + BiggerUINT32
        solvers[solver][
            "fault_description"] = "Return Code of MaxSAT solver == 255 (SIGTERM / SIGABORT)"
    elif solvers[solver].get('return_code', 0) != 0 and solvers[solver].get('return_code', 0) != 10 and solvers[solver].get('return_code', 0) != 20 and solvers[solver].get('return_code', 0) != 30:
        solvers[solver]["error_code"] = 5 + BiggerUINT32
        solvers[solver]["fault_description"] = "Invalid Return Code of MaxSAT solver == " + \
            str(solvers[solver].get('return_code', 0))
    elif minVerifiedOValue == 2**64 and verifiedUNSAT == "SATISFIABLE":
        solvers[solver]["error_code"] = 16 + BiggerUINT32
        solvers[solver]["fault_description"] = "All solver have a wrong result, as verifier gives SATISFIABLE but no solver found a solution which is verifiable."
    elif solvers[solver].get('ERROR_solver', '') == "opt_but_no_v-model":
        solvers[solver]["error_code"] = 6 + BiggerUINT32
        solvers[solver]["fault_description"] = "s return status is OPTIMUM FOUND - but no model given"
    elif solvers[solver].get('ERROR_solver', '') == "opt_but_no_o-value":
        solvers[solver]["error_code"] = 7 + BiggerUINT32
        solvers[solver]["fault_description"] = "s return status is OPTIMUM FOUND - but no o value given"
    elif solvers[solver]["s-status"] == "UNKNOWN":
        solvers[solver]["error_code"] = 23 + BiggerUINT32
        solvers[solver]["fault_description"] = "s UNKNOWN status line in solver output."
    elif solvers[solver].get('ERROR_solver', '') == "v-model_and_o-value_but_no_opt":
        solvers[solver]["error_code"] = 8 + BiggerUINT32
        solvers[solver]["fault_description"] = "model and o value given but no solver return status is not OPTIMUM FOUND"
    elif solvers[solver].get("ver_return_code", 0) == 1:
        solvers[solver]["error_code"] = 9 + BiggerUINT32
        solvers[solver]["fault_description"] = "Verifier returned, that given model is UNSAT"
    elif solvers[solver]["s-status"] == "UNSATISFIABLE" and verifiedUNSAT == "SATISFIABLE":
        solvers[solver]["error_code"] = 10 + BiggerUINT32
        solvers[solver]["fault_description"] = "Verifier returned SATISFIABLE and solver returned UNSATISFIABLE"
    elif verifiedUNSAT != "UNSATISFIABLE" and verifiedUNSAT != "SATISFIABLE":
        solvers[solver]["error_code"] = 11 + BiggerUINT32
        solvers[solver]["fault_description"] = "STRANGE RESULT of SAT Solver in verifier - " + \
            str(verifiedUNSAT)
    elif solvers[solver].get("o-value", -1) != solvers[solver].get('ver_o-value', -1) and solvers[solver].get('ver_o-value', -1) != minVerifiedOValue and solvers[solver].get("o-value", -1) != minVerifiedOValue:
        solvers[solver]["error_code"] = 12 + BiggerUINT32
        solvers[solver]["fault_description"] = "MaxSAT solver the o-values given by the solver, by the verifier of the model and the minimal verified o value are three different values."
    elif solvers[solver].get("o-value", -1) == minVerifiedOValue and solvers[solver].get("ver_o-value", -1) > minVerifiedOValue:
        solvers[solver]["error_code"] = 13 + BiggerUINT32
        solvers[solver]["fault_description"] = "MaxSAT o value equals minimal verified o value BUT given model has a bigger o-value."
    elif solvers[solver].get("o-value", -1) > solvers[solver].get('ver_o-value', -1) and solvers[solver].get('ver_o-value', -1) == minVerifiedOValue:
        solvers[solver]["error_code"] = 14 + BiggerUINT32
        solvers[solver]["fault_description"] = "MaxSAT solver o value is bigger than the o-value of its model which equals the minimal verified o-value."
    elif solvers[solver].get("o-value", 2**64) < solvers[solver].get('ver_o-value', -1) and solvers[solver].get('ver_o-value', -1) == minVerifiedOValue:
        solvers[solver]["error_code"] = 15 + BiggerUINT32
        solvers[solver]["fault_description"] = "MaxSAT solver o value is smaller than the o-value of its model, but the o-value of it's model  equals the minimal verified o-value.."
    # SHOULD NEVER OCCUR as this error is caught by 12 and 14
    # elif solvers[solver].get("o-value", -1) > solvers[solver].get('ver_o-value', 2**64):
    #     solvers[solver]["error_code"] = 16 + BiggerUINT32
    #     solvers[solver]["fault_description"] = "MaxSAT solver o-value is bigger than the o-value of it's model."
    # SHOULD NEVER OCCUR as this error is caught by 12 and 15
    # elif solvers[solver].get("o-value", 2**64) < solvers[solver].get('ver_o-value', -1):
    #     solvers[solver]["error_code"] = 17 + BiggerUINT32
    #     solvers[solver]["fault_description"] = "MaxSAT solver o-value is smaller than the o-value of it's model."
    elif solvers[solver].get("ver_o-value", -1) > minVerifiedOValue:
        solvers[solver]["error_code"] = 18 + BiggerUINT32
        solvers[solver]["fault_description"] = "o-value of MaxSAT solver model is bigger than the minimal verified o-value."
    elif solvers[solver].get("o-value", -1) > minVerifiedOValue:
        solvers[solver]["error_code"] = 19 + BiggerUINT32
        solvers[solver]["fault_description"] = "o-value of MaxSAT solver is bigger than the minimal verified o-value."
    elif solvers[solver].get("o-value", 2**64) < minVerifiedOValue:
        solvers[solver]["error_code"] = 20 + BiggerUINT32
        solvers[solver]["fault_description"] = "o-value of MaxSAT solver is smaller than the minimal verified o-value."
    elif solvers[solver].get("o-value", -1) != -1 and solvers[solver].get("ver_o-value", -1) != -1 and solvers[solver].get("o-value", -1) != solvers[solver].get("ver_o-value", -1) or solvers[solver].get("o-value", 2**64) != minVerifiedOValue:
        solvers[solver]["error_code"] = 21 + BiggerUINT32
        solvers[solver]["fault_description"] = "o-value is unequal to the o-value of model  OR  o-value is unequal of minimal verified o-value."
    elif solvers[solver].get("ver_return_code", 0) != 0 and solvers[solver].get("ver_return_code", -1) != 1 and solvers[solver].get("ver_return_code", -1) != 2 and solvers[solver].get("ver_return_code", -1) != 3 and solvers[solver].get("ver_return_code", -1) != 11:
        solvers[solver]["error_code"] = 22 + BiggerUINT32
        solvers[solver]["fault_description"] = "INVALID return code " + \
            str(solvers[solver].get("ver_return_code", 0)) + \
            " of the model verifier!!!"
    elif solvers[solver]["s-status"] == "STRANGE":
        solvers[solver]["error_code"] = 24 + BiggerUINT32
        solvers[solver]["fault_description"] = "STRANGE value in solver status output."
    ### NUMBERS ARE USED NOW DIFFERNTLY!!!
    # SHOULD NEVER OCCUR as in this case this is catched earlier
    # elif solvers[solver]["ver_result"] == "FALSE VALUE":
    #     solvers[solver]["error_code"] =  + BiggerUINT32
    #     solvers[solver]["fault_description"] = "Verifier claims that model and o-value do not match."
    # SHOULD NEVER OCCUR as in this case the verifier does not verify the given model and this is catched earlier
    # elif solvers[solver]["ver_result"] == "UNSAT BUT VALUE":
    #     solvers[solver]["error_code"] =  + BiggerUINT32
    #     solvers[solver]["fault_description"] = "Verifier claims that the hard clauses are UNSAT but a o value is given."
    # SHOULD NEVER OCCUR as it should be caught by 10
    # elif solvers[solver]["ver_result"] == "FALSE UNSAT":
    #     solvers[solver]["error_code"] =  + BiggerUINT32
    #     solvers[solver]["fault_description"] = "Verifier claims that the hard clauses are SAT but MaxSAT solver tells UNSAT."
    elif solvers[solver]["ver_result"] == "UNKNOWN":
        solvers[solver]["error_code"] = 28 + BiggerUINT32
        solvers[solver]["fault_description"] = "UNKNOWN verifier result"
    elif solvers[solver].get('stderr', '') != '':
        solvers[solver]["error_code"] = 29 + BiggerUINT32
        solvers[solver]["fault_description"] = "MaxSAT solver returned something in stderr."
    elif solvers[solver].get('stdout_error', '') != '':
        solvers[solver]["error_code"] = 30 + BiggerUINT32
        solvers[solver]["fault_description"] = "MaxSAT solver had ERROR written in some form in stdout"
    else:
        solvers[solver]["error_code"] = 0

    if solvers[solver]["error_code"] != 0 and solver in args.solvers:
        overall_return_code += solvers[solver]["id"] + \
            solvers[solver].get('error_code', 0)
        if args.logging or args.logAll:
            logging_files.append(args.logPath + "/" + solver + "_" +
                                 str(solvers[solver]["error_code"]) + "." + str(long_seed) + ".lg")
    elif solvers[solver]["error_code"] != 0:
        if args.logAll:
            logging_files.append(args.logPath + "/" + solver + "_" +
                                 str(solvers[solver]["error_code"]) + "." + str(long_seed) + ".lg")
    if solvers[solver]["error_code"] != 0:
        faulty_solvers.append(solver)

while (overall_return_code > 255):
    overall_return_code = round(overall_return_code/2 + 0.5)


if len(logging_files) > 0:
    print("c Create the following logfile(s): " + str(logging_files))

files = [open(file, 'w') for file in logging_files]


def Logging(string, fd_list=files, to_console=True, prefix=''):

    for fd in fd_list:
        for line in string.split("\n"):
            fd.write(prefix + line + "\n")
    if to_console:
        print(prefix + string)


Logging("c WCNF-FILE....: " + args.wcnfFile)

destFile = ""
if len(faulty_solvers) > 0 and args.saveWCNF:
    if not os.path.exists(args.saveWCNFFolder):
        os.makedirs(args.saveWCNFFolder)
    if destFile == "":
        destFile = args.saveWCNFFolder + "/" + os.path.basename(args.wcnfFile)[:-5] + "-" + str(random.randrange(1, 2**32-1)) + ".wcnf"
        shutil.copy(args.wcnfFile, destFile)
    Logging("c copy saved in: " + args.saveWCNFFolder)

Logging("c SEED.........: " + seed)
Logging(instance_stats)

if BiggerUINT32 == 0:
    Logging("c SumOfWeights.: < UINT32")
else:
    Logging("c SumOfWeights.: > UINT32")
Logging("c Hard clauses.: " + str(verifiedUNSAT))
Logging("c solver times.: " + str(totalTime) + "  (without timeouts)")
Logging("c " + "name".ljust(22) + "error_code".rjust(12) + "solver_rc".rjust(10) + "ver_result".rjust(22) +
        "o-value".rjust(22) + "ver_o-value".rjust(22) + "min_ver_o-value".rjust(22) + "solver_error".rjust(22))

if (minVerifiedOValue == 2**64):
    minVeri = "-"
else:
    minVeri = minVerifiedOValue


for solver in solvers:
    if solvers[solver].get('error_code', 0) != 0:
        current_error = '!' + \
            str(solvers[solver].get('error_code', 0))
    else:
        current_error = "0"

    Logging("c " + solvers[solver].get('name', '').ljust(22) + current_error.rjust(12) + str(solvers[solver].get('return_code', 0)).rjust(10) + solvers[solver].get('ver_result', '').rjust(22) +
            str(solvers[solver].get('o-value', '-')).rjust(22) + str(solvers[solver].get('ver_o-value', '-')).rjust(22) + str(minVeri).rjust(22) + solvers[solver].get('ERROR_solver', '-').rjust(22))

Logging("c faulty solvers:: " + str(faulty_solvers))

for solver in faulty_solvers:
    if solvers[solver].get("fault_description", "") != "":
        Logging("c FAULT DESCRIPTION of " + solver + ": " +
                solvers[solver].get("fault_description", ""))

if args.logging:
    for solver, file in zip(faulty_solvers, files):
        if solvers[solver].get("stdout", '') != '':
            Logging(str(solvers[solver].get("stdout", '')), [
                    file], False, "[STDOUT" + "] ")
        if solvers[solver].get("stderr", '') != '':
            Logging(str(solvers[solver].get("stderr", '')), [
                    file], False, "[STDERR" + "] ")


Logging("c rc:" + str(overall_return_code))

[file.close() for file in files]
exit(overall_return_code)
