const char *usage =
    "usage: ddwcnf [ <option> ... ] <wdimacs>\n"
    "\n"
    "where '<option>' can be one of the following\n"
    "\n"
    "  -h | --help                 print this command line option summary\n\n"
    "  -s | --satSolver            SAT solver to verify UNSAT of hard clauses\n"
    "                              if a SAT solver is given only UNSAT is "
    "checked\n"
    "  -o | --expectedMin          expected o value to check against\n"
    "(v-line) of a MaxSAT solver\n"
    "  -v | --result               v-line or resulting model, file must be given\n"
    "  -c | --convert              convert to old DIMACS format, file must be given\n"

#ifdef LOGGING
    "  -l | --logging              print very verbose logging information\n"
    "                              using -l multiple times increases the "
    "verbosity\n"
#endif
    //"  -s | --static      static decision order\n"
    "\n"
    "and '<wdimacs>' is the input file in WDIMACS format.\n";
// NOT TESTED!! -- "The solver reads from '<stdin>' if no input file is
// specified.\n";

#include <cstdarg>
#include <cstdio>
#include <zlib.h> // parser

#include <climits>
#include <cstdlib>
#include <sys/resource.h>
#include <sys/time.h>
// #include <cstring> // fast parser
#include <algorithm> // count_if
#include <fstream>   // parseWcnfFile std::ifstream
#include <random>    // shuffle std::vector
#include <set>
#include <sstream> // parseWcnfFile std::istringstream
#include <string>
#include <unordered_map>

#include <cassert> // assertions
#include <cstring>
#include <iomanip>  //std::setw
#include <iostream> //std::cout
#include <vector>

// Solver data.
struct clause {
  size_t size;
  uint64_t weight;
  bool hard;
  bool active;

  // Flexible Array Member(FAM)
  int literals[];

  int *begin() { return literals; }
  int *end() { return literals + size; }
};
std::vector<clause *> clauses;

const char *fileName;       // given file name
std::string tmpCNFFileName; // temporary cnf file
std::string tmpWCNFFileName; // temporary wcnf file
std::string solver;

int nbClauses = 0;
int nbSoftClauses = 0;
int nbVars = 0;
int nbVarsInHard = 0;
uint64_t sumOfSoftWeights = 0;
uint64_t maxWeight = 0;

void reset(void) {
  for (auto clause : clauses) {
    delete[] clause;
  }
}

static void die(const char *fmt, ...) {
  fprintf(stderr, "c solver error: ");
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  fputc('\n', stderr);
  reset();
  exit(1);
}

#ifdef LOGGING
unsigned logging = 0;

void debug(clause *clause, unsigned loglevel) {
  if (loglevel > logging)
    return;
  std::cout << "c DEBUG CLAUSE:  ";
  for (auto lit : *clause) {
    std::cout << lit << ", ";
  }
  std::cout << std::endl;
}

#define dout0 std::cout << "c DEBUG (" << __LINE__ << "): "
#define dout1                                                                  \
  if (logging > 0)                                                             \
  std::cout << "c DEBUG (" << __LINE__ << "): "
#define dout2                                                                  \
  if (logging > 1)                                                             \
  std::cout << "c DEBUG (" << __LINE__ << "): "
#define dout3                                                                  \
  if (logging > 2)                                                             \
  std::cout << "c DEBUG (" << __LINE__ << "): "

#else
#define dout0 0 && std::cout
#define dout1 0 && std::cout
#define dout2 0 && std::cout
#define dout3 0 && std::cout
#define debug(...)                                                             \
  do {                                                                         \
  } while (0)
#endif

//-------------------------------------------------------------------------------------------------
// A simple buffered character stream class:
// static const int buffer_size = 1048576;

class StreamBuffer {
  gzFile in;
  unsigned char buf[1048576];
  int pos;
  int size;

  void assureLookahead() {
    if (pos >= size) {
      pos = 0;
      size = gzread(in, buf, sizeof(buf));
    }
  }

public:
  explicit StreamBuffer(gzFile i) : in(i), pos(0), size(0) {
    assureLookahead();
  }

  int operator*() const { return (pos >= size) ? EOF : buf[pos]; }
  void operator++() {
    pos++;
    assureLookahead();
  }
  int position() const { return pos; }
};

//-------------------------------------------------------------------------------------------------
// End-of-file detection functions for StreamBuffer and char*:

static inline bool isEof(StreamBuffer &in) { return *in == EOF; }
static inline bool isEof(const char *in) { return *in == '\0'; }

//-------------------------------------------------------------------------------------------------
// Generic parse functions parametrized over the input-stream type.

template <class B> static void skipWhitespace(B &in) {
  while ((*in >= 9 && *in <= 13) || *in == 32)
    ++in;
}

template <class B> static void skipLine(B &in) {
  for (;;) {
    if (isEof(in))
      return;
    if (*in == '\n') {
      ++in;
      return;
    }
    ++in;
  }
}

template <class B> static int parseInt(B &in) {
  int val = 0;
  bool neg = false;
  skipWhitespace(in);

  if (*in == '-')
    neg = true, ++in;
  else if (*in == '+')
    ++in;
  if (*in < '0' || *in > '9') {
    std::cerr << "PARSE ERROR!!! Unexpected char in parseInt: " << *in
              << std::endl;
    exit(3);
  }
  while (*in >= '0' && *in <= '9') {
    val = val * 10 + (*in - '0');
    ++in;
  }
  assert(val != 0);
  return neg ? -val : val;
}

template <class B> static uint64_t parseWeight(B &in) {

  uint64_t val = 0;
  skipWhitespace(in);
  if (*in == 'h') {
    ++in;
    if (*in != ' ') {
      std::cerr << "o PARSE ERROR! Unexpected char in parseWeight: " << *in
                << std::endl;
      exit(3);
    }
    return UINT64_MAX;
  }
  if (*in == '-') {
    std::cerr << "o PARSE ERROR! Unexpected negative weight" << std::endl;
    exit(3);
  } else if (*in == '+')
    ++in;
  if (*in < '0' || *in > '9') {
    std::cerr << "o PARSE ERROR! Unexpected char in parseWeight: " << *in
              << std::endl;
    exit(3);
  }
  while (*in >= '0' && *in <= '9')
    val = val * 10 + (*in - '0'), ++in;
  return val;
}

// String matching: consumes characters eagerly, but does not require random
// access iterator.
template <class B> static bool eagerMatch(B &in, const char *str) {
  for (; *str != '\0'; ++str, ++in)
    if (*str != *in)
      return false;
  return true;
}

void add_new_clause(const std::vector<int> &literals, uint64_t &weight) {
  size_t size = literals.size();
  size_t bytes = sizeof(struct clause) + size * sizeof(int);
  clause *cl = (clause *)new char[bytes];

  cl->size = size;
  cl->active = true;

  for (unsigned i = 0; i < literals.size(); i++)
    cl->literals[i] = literals[i];

  debug(cl, 3);
  cl->weight = weight;
  cl->hard = (weight == UINT64_MAX);

  assert(clauses.size() <= (size_t)INT_MAX);
  clauses.push_back(cl);
}

void parseWCNF(std::string wcnfFile) {
  dout2 << __PRETTY_FUNCTION__ << std::endl;

  gzFile input_stream = gzopen(wcnfFile.c_str(), "rb");
  if (input_stream == nullptr) {
    std::cout << "c ERROR! Could not open file: " << wcnfFile << std::endl;
    return;
  }
  StreamBuffer in(input_stream);
  std::vector<int> clause;
  uint64_t weight;

  while (true) {
    if (isEof(in)) {
      break;
    } else if (*in == 'c' || *in == '\n') {
      skipLine(in);
      continue;
    }

    weight = parseWeight(in);
    if (weight != UINT64_MAX) {
      if (weight > maxWeight)
        maxWeight = weight;
      sumOfSoftWeights += weight;
      nbSoftClauses++;
    } else
      nbClauses++;
    clause.clear();
    while (true) {
      skipWhitespace(in);
      if (*in == '0') {
        skipLine(in);
        break;
      }
      clause.push_back(parseInt(in));
      if (abs(clause.back()) > nbVars)
        nbVars = abs(clause.back());
      if (weight == UINT64_MAX)
        if (abs(clause.back()) > nbVarsInHard)
          nbVarsInHard = abs(clause.back());
    }

    assert(clause.size() > 0);
    add_new_clause(clause, weight);
  }
  gzclose(input_stream);
}

void dumpHardWCNF() {
  std::cout << "p cnf " << nbVarsInHard << " " << nbClauses << std::endl;
  for (auto *clause : clauses) {
    if (!clause->hard)
      continue;
    for (auto lit : *clause) {
      std::cout << lit << " ";
    }
    std::cout << "0" << std::endl;
  }
}

void writeCNF() {
  dout2 << __PRETTY_FUNCTION__ << std::endl;
  dout2 << "Write cnf file: " << tmpCNFFileName << std::endl;
  // Save the rdbuf of cout.
  std::ofstream out(tmpCNFFileName.c_str());
  auto *coutbuf = std::cout.rdbuf(); // save old buf
  std::cout.rdbuf(out.rdbuf());      // redirect std::cout to tmpWCNFFileName!
  dumpHardWCNF();
  std::cout.rdbuf(coutbuf);
  // dumpHardWCNF();
}

void ConvertToOldWCNFFormat() {
  if (sumOfSoftWeights == UINT64_MAX) {
    sumOfSoftWeights--;
  }
  std::cout << "p wcnf " << nbVars << " " << clauses.size() << " "
            << sumOfSoftWeights + 1 << std::endl;
  for (auto clause : clauses) {
    if (clause->hard)
      std::cout << sumOfSoftWeights + 1 << " ";
    else
      std::cout << clause->weight << " ";
    for (auto literal : *clause) {
      std::cout << literal << " ";
    }
    std::cout << "0" << std::endl;
  }
}

void writeOldWCNF() {
  dout2 << __PRETTY_FUNCTION__ << std::endl;
  dout2 << "Write wcnf file: " << tmpWCNFFileName << std::endl;
  // Save the rdbuf of cout.
  std::ofstream out(tmpWCNFFileName.c_str());
  auto *coutbuf = std::cout.rdbuf(); // save old buf
  std::cout.rdbuf(out.rdbuf());      // redirect std::cout to tmpWCNFFileName!
  ConvertToOldWCNFFormat();
  std::cout.rdbuf(coutbuf);
  // dumpHardWCNF();
}

int hardClausesUNSAT() {
  dout2 << __PRETTY_FUNCTION__ << std::endl;
  writeCNF();
  // int rv = WEXITSTATUS(std::system((solver + " hard.cnf > /dev/null
  // 2>&1").c_str()));
  int rv = WEXITSTATUS(std::system(
      (solver + " " + tmpCNFFileName + " > /dev/null 2>&1").c_str()));
  if (WEXITSTATUS(std::system(("rm " + tmpCNFFileName).c_str()))) {
    std::cerr << "could not remove hard.cnf" << std::endl;
  }

  if (rv == 20) {
    std::cout << "c hard clauses are UNSATISFIABLE" << std::endl;
    return 0;
  } else if (rv == 10) {
    std::cout << "c hard clauses are SATISFIABLE" << std::endl;
    return 11;
  } else if (rv == 0) {
    std::cout << "c " << solver << " returns UNKNOWN" << std::endl;
    return 12;
  } else {
    assert(false);
    std::cerr << "c unexpected return value of " << solver << ": " << rv
              << std::endl;
    return 42;
  }
  return 42;
}

int checkResult(uint64_t ovalue, std::string modelFile) {

  dout2 << __PRETTY_FUNCTION__ << std::endl;

  std::string model = "";
  gzFile input_stream = gzopen(modelFile.c_str(), "rb");
  if (input_stream == nullptr) {
    std::cout << "c ERROR! Could not open model file: " << model << std::endl;
    return -1;
  }
  StreamBuffer in(input_stream);

  while (true) {
    if (isEof(in)) {
      break;
    } else if (*in != 'v' || *in == '\n') {
      skipLine(in);
      continue;
    }
    ++in;               // skip letter v
    skipWhitespace(in); // skip whitespace
    while (*in != '\n' && (*in == '0' || *in == '1')) {
      // if (*in != '0' && *in != '1')
      //   std::cout << *in << " invalid  *in" << std::endl;
      model += *in;
      ++in;
    }
    // assert(*in == '\n'); -- not necessarily, if it is a one liner...
  }
  dout0 << "v " << model << std::endl;

  uint64_t sumWeights = 0;
  assert(ovalue != UINT64_MAX);
  assert(model != "");
  for (auto clause : clauses) {
    bool sat = false;
    for (auto lit : *clause) {
      assert(model.size() >= abs(lit));
      assert(0 < abs(lit));
      bool value = (model[abs(lit) - 1] == '1');
      if ((lit > 0 && value) || (lit < 0 && !value)) {
        sat = true;
        break;
      }
    }
    if (!sat && clause->hard) {
      dout0 << "a hard clause is not satisfiable" << std::endl;
      return 1;
    } else if (!sat && !clause->hard)
      sumWeights += clause->weight;
  }
  std::cout << "o " << sumWeights << std::endl;
  if (ovalue < sumWeights) {
    dout0 << "sumWeights: " << sumWeights << "  ovalue: " << ovalue
          << std::endl;
    return 2;
  } else if (ovalue > sumWeights) {
    dout0 << "sumWeights: " << sumWeights << "  ovalue: " << ovalue
          << std::endl;
    return 3;
  }
  return 0;
}

int main(int argc, char **argv) {

  dout2 << __PRETTY_FUNCTION__ << std::endl;
  // double timeStart;
  // struct rusage resources;
  // getrusage(RUSAGE_SELF, &resources);
  // timeStart =
  //     resources.ru_utime.tv_sec + 1.e-6 * (double)resources.ru_utime.tv_usec;
  std::cout << "c Fancy WDIMACS Result Checker." << std::endl;
  bool unsat = false;
  uint64_t ovalue = UINT64_MAX;
  std::string model = "";
  unsigned seed = abs((int)(time(NULL) * getpid()) >> 1);
  srand(seed);
  int rndPrefix = rand();
  // std::cout << "c random prefix.......: " << rndPrefix << std::endl;

  for (int i = 1; i != argc; i++) {
    const char *arg = argv[i];
    if (!strcmp(arg, "-h") || !strcmp(arg, "--help")) {
      fputs(usage, stdout);
      exit(0);
    } else if (!strcmp(arg, "-s") || !strcmp(arg, "--satSolver")) {
      solver = argv[++i];
      unsat = true;
    } else if (!strcmp(arg, "-o") || !strcmp(arg, "--expectedMin")) {
      ovalue = strtoull(argv[++i], NULL, 0);
    } else if (!strcmp(arg, "-v") || !strcmp(arg, "--result")) {
      model = argv[++i];
    } else if (!strcmp(arg, "-c") || !strcmp(arg, "--convert")) {
      tmpWCNFFileName = argv[++i];
    } else if (!strcmp(arg, "-l") || !strcmp(arg, "--logging")) {
#ifdef LOGGING
      logging++;
#else
      die("compiled without logging code (use './configure --logging')");
      exit(1);
#endif
    } else if (arg[0] == '-')
      die("invalid option '%s' (try '-h')", arg);
    else if (fileName)
      die("too many arguments '%s' and '%s' (try '-h')", fileName, arg);
    else
      fileName = arg;
  }

  if (!fileName) {
    fileName = "<stdin>";
  }

  parseWCNF(fileName);

  if (tmpWCNFFileName != "") {
    writeOldWCNF();
    reset();
    exit(0);
  }

  int rv = 0;
  if (unsat && solver != "") {
    if (nbVarsInHard > 0) {
      tmpCNFFileName = "/tmp/" + std::to_string(rndPrefix) + ".cnf";
      rv = hardClausesUNSAT();
    } else
      rv = 11;
    std::cout << "c Clauses......: " << nbClauses << std::endl;
    std::cout << "c SoftClauses..: " << nbSoftClauses << std::endl;
    std::cout << "c Variables....: " << nbVars << std::endl;
    std::cout << "c MaxWeight....: " << maxWeight << std::endl;
    std::cout << "c SumOfWeights.: " << sumOfSoftWeights << std::endl;
  } else if (ovalue != UINT64_MAX && model != "") {
    rv = checkResult(ovalue, model);
  } else {
    rv = 42;
    assert(false);
  }

  reset();
  // getrusage(RUSAGE_SELF, &resources);

  // std::cout << "time.........: " << timeStart - (resources.ru_utime.tv_sec
  // + 1.e-6 * (double)resources.ru_utime.tv_usec) << std::endl;
  return rv;
}
